@@ -5,20 +5,16 @@ import struct
 import sys
 
 from peewee import *
-from peewee import Alias
+from peewee import ColumnBase
 from peewee import EnclosedNodeList
 from peewee import Entity
 from peewee import Expression
-from peewee import Function
 from peewee import Node
 from peewee import NodeList
 from peewee import OP
-from peewee import Select
 from peewee import VirtualField
 from peewee import merge_dict
-from peewee import sort_models
 from peewee import sqlite3
-from peewee import text_type
 try:
     from playhouse._sqlite_ext import (
         backup,
@@ -73,23 +69,50 @@ class AutoIncrementField(AutoField):
         return NodeList((node_list, SQL('AUTOINCREMENT')))
 
 
-class JSONPath(Node):
-    def __init__(self, path=None):
-        self._path = path or []
+class JSONPath(ColumnBase):
+    def __init__(self, field, path=None):
+        super(JSONPath, self).__init__()
+        self._field = field
+        self._path = path or ()
+
+    @property
+    def path(self):
+        return Value('$%s' % ''.join(self._path))
 
     def __getitem__(self, idx):
-        path = list(self._path)
         if isinstance(idx, int):
-            path.append('[%s]' % idx)
+            item = '[%s]' % idx
         else:
-            path.append('.%s' % idx)
-        return JSONPath(path)
-    __getattr__ = __getitem__
+            item = '.%s' % idx
+        return JSONPath(self._field, self._path + (item,))
+
+    def set(self, value, as_json=None):
+        if as_json or isinstance(value, (list, dict)):
+            value = fn.json(json.dumps(value))
+        return fn.json_set(self._field, self.path, value)
+
+    def update(self, value):
+        return self.set(fn.json_patch(self, json.dumps(value)))
+
+    def remove(self):
+        return fn.json_remove(self._field, self.path)
+
+    def json_type(self):
+        return fn.json_type(self._field, self.path)
+
+    def length(self):
+        return fn.json_array_length(self._field, self.path)
+
+    def children(self):
+        return fn.json_each(self._field, self.path)
+
+    def tree(self):
+        return fn.json_tree(self._field, self.path)
 
     def __sql__(self, ctx):
-        return ctx.sql(Value('$%s' % ''.join(self._path)))
+        return ctx.sql(fn.json_extract(self._field, self.path)
+                       if self._path else self._field)
 
-J = JSONPath()
 
 class JSONField(TextField):
     field_type = 'JSON'
@@ -105,76 +128,25 @@ class JSONField(TextField):
         if value is not None:
             return json.dumps(value)
 
-    def clean_path(self, path):
-        if isinstance(path, JSONPath):
-            return path
-        if path.startswith('[') or not path:
-            return '$%s' % path
-        return '$.%s' % path
+    def __getitem__(self, item):
+        return JSONPath(self)[item]
 
-    def clean_paths(self, paths):
-        return [self.clean_path(path) for path in paths]
-
-    def length(self, *paths):
-        if paths:
-            return fn.json_array_length(self, *self.clean_paths(paths))
-        return fn.json_array_length(self)
-
-    def extract(self, *paths):
-        return fn.json_extract(self, *self.clean_paths(paths))
-
-    def __getitem__(self, path):
-        if not isinstance(path, tuple):
-            path = (path,)
-        return self.extract(*path)
-
-    def _value_for_insertion(self, value):
-        if isinstance(value, (list, tuple, dict)):
-            return fn.json(json.dumps(value))
-        return value
-
-    def _insert_like(self, fn, pairs, mapping):
-        npairs = len(pairs)
-        if npairs == 1 and isinstance(pairs[0], dict):
-            # We were passed a single dictionary to update the field with.
-            mapping.update(pairs[0])
-            npairs = 0
-        elif npairs % 2 != 0:
-            raise ValueError('Unequal key and value parameters.')
-
-        accum = []
-        if npairs:
-            for i in range(0, npairs, 2):
-                accum.append(self.clean_path(pairs[i]))
-                accum.append(self._value_for_insertion(pairs[i + 1]))
-        if mapping:
-            for key in mapping:
-                val = mapping[key]
-                accum.append(self.clean_path(key))
-                accum.append(self._value_for_insertion(val))
-        return fn(self, *accum)
-
-    def insert(self, *pairs, **data):
-        return self._insert_like(fn.json_insert, pairs, data)
-
-    def replace(self, *pairs, **data):
-        return self._insert_like(fn.json_replace, pairs, data)
-
-    def set(self, *pairs, **data):
-        return self._insert_like(fn.json_set, pairs, data)
-
-    def remove(self, *paths):
-        return fn.json_remove(self, *self.clean_paths(paths))
+    def set(self, value, as_json=None):
+        return JSONPath(self).set(value, as_json)
 
     def update(self, data):
-        return fn.json_patch(self, self._value_for_insertion(data))
+        return JSONPath(self).update(data)
 
-    def json_type(self, path=None):
-        if path:
-            return fn.json_type(self, self.clean_path(path))
+    def remove(self):
+        return JSONPath(self).remove()
+
+    def json_type(self):
         return fn.json_type(self)
 
-    def children(self, path=None):
+    def length(self):
+        return fn.json_array_length(self)
+
+    def children(self):
         """
         Schema of `json_each` and `json_tree`:
 
@@ -189,13 +161,9 @@ class JSONField(TextField):
         json JSON hidden (1st input parameter to function)
         root TEXT hidden (2nd input parameter, path at which to start)
         """
-        if path:
-            return fn.json_each(self, self.clean_path(path))
         return fn.json_each(self)
 
-    def tree(self, path=None):
-        if path:
-            return fn.json_tree(self, self.clean_path(path))
+    def tree(self):
         return fn.json_tree(self)
 
 
@@ -207,6 +175,9 @@ class SearchField(Field):
         super(SearchField, self).__init__(unindexed=unindexed,
                                           column_name=column_name, null=True)
 
+    def match(self, term):
+        return match(self, term)
+
 
 class VirtualTableSchemaManager(SchemaManager):
     def _create_virtual_table(self, safe=True, **options):
@@ -602,10 +573,7 @@ class FTS5Model(BaseFTSModel):
 
     @classmethod
     def rank(cls, *args):
-        if args:
-            return cls.bm25(*args)
-        else:
-            return SQL('rank')
+        return cls.bm25(*args) if args else SQL('rank')
 
     @classmethod
     def bm25(cls, *weights):
@@ -651,7 +619,7 @@ class FTS5Model(BaseFTSModel):
                 .order_by(order_by))
 
     @classmethod
-    def _fts_cmd(cls, cmd, **extra_params):
+    def _fts_cmd_sql(cls, cmd, **extra_params):
         tbl = cls._meta.entity
         columns = [tbl]
         values = [cmd]
@@ -659,14 +627,17 @@ class FTS5Model(BaseFTSModel):
             columns.append(Entity(key))
             values.append(value)
 
-        inner_clause = EnclosedClause(tbl)
-        clause = Clause(
+        return NodeList((
             SQL('INSERT INTO'),
             cls._meta.entity,
-            EnclosedClause(*columns),
+            EnclosedNodeList(columns),
             SQL('VALUES'),
-            EnclosedClause(*values))
-        return cls._meta.database.execute(clause)
+            EnclosedNodeList(values)))
+
+    @classmethod
+    def _fts_cmd(cls, cmd, **extra_params):
+        query = cls._fts_cmd_sql(cmd, **extra_params)
+        return cls._meta.database.execute(query)
 
     @classmethod
     def automerge(cls, level):
@@ -924,7 +895,7 @@ class LSMTable(VirtualModel):
 OP.MATCH = 'MATCH'
 
 def _sqlite_regexp(regex, value):
-    return re.search(regex, value, re.I) is not None
+    return re.search(regex, value) is not None
 
 
 class SqliteExtDatabase(SqliteDatabase):
@@ -998,8 +969,8 @@ if CYTHON_SQLITE_EXTENSIONS:
 
     def __status__(flag, return_highwater=False):
         """
-        Expose a sqlite3_status() call for a particular flag as a property of the
-        Database object.
+        Expose a sqlite3_status() call for a particular flag as a property of
+        the Database object.
         """
         def getter(self):
             result = sqlite_get_status(flag)
@@ -1008,11 +979,13 @@ if CYTHON_SQLITE_EXTENSIONS:
 
     def __dbstatus__(flag, return_highwater=False, return_current=False):
         """
-        Expose a sqlite3_dbstatus() call for a particular flag as a property of the
-        Database instance. Unlike sqlite3_status(), the dbstatus properties pertain
-        to the current connection.
+        Expose a sqlite3_dbstatus() call for a particular flag as a property of
+        the Database instance. Unlike sqlite3_status(), the dbstatus properties
+        pertain to the current connection.
         """
         def getter(self):
+            if self._state.conn is None:
+                raise ImproperlyConfigured('database connection not opened.')
             result = sqlite_get_db_status(self._state.conn, flag)
             if return_current:
                 return result[0]
@@ -1049,7 +1022,8 @@ if CYTHON_SQLITE_EXTENSIONS:
             if self._update_hook is not None:
                 self._conn_helper.set_update_hook(self._update_hook)
             if self._replace_busy_handler:
-                self._conn_helper.set_busy_handler(self.timeout or 5000)
+                timeout = self._timeout or 5
+                self._conn_helper.set_busy_handler(timeout * 1000)
 
         def on_commit(self, fn):
             self._commit_hook = fn
@@ -1080,11 +1054,14 @@ if CYTHON_SQLITE_EXTENSIONS:
         def autocommit(self):
             return self._conn_helper.autocommit()
 
-        def backup(self, destination):
-            return backup(self.connection(), destination.connection())
+        def backup(self, destination, pages=None, name=None, progress=None):
+            return backup(self.connection(), destination.connection(),
+                          pages=pages, name=name, progress=progress)
 
-        def backup_to_file(self, filename):
-            return backup_to_file(self.connection(), filename)
+        def backup_to_file(self, filename, pages=None, name=None,
+                           progress=None):
+            return backup_to_file(self.connection(), filename, pages=pages,
+                                  name=name, progress=progress)
 
         def blob_open(self, table, column, rowid, read_only=False):
             return Blob(self, table, column, rowid, read_only)
@@ -1125,6 +1102,15 @@ def _parse_match_info(buf):
     bufsize = len(buf)  # Length in bytes.
     return [struct.unpack('@I', buf[i:i+4])[0] for i in range(0, bufsize, 4)]
 
+def get_weights(ncol, raw_weights):
+    if not raw_weights:
+        return [1] * ncol
+    else:
+        weights = [0] * ncol
+        for i, weight in enumerate(raw_weights):
+            weights[i] = weight
+    return weights
+
 # Ranking implementation, which parse matchinfo.
 def rank(raw_match_info, *raw_weights):
     # Handle match_info called w/default args 'pcx' - based on the example rank
@@ -1133,13 +1119,14 @@ def rank(raw_match_info, *raw_weights):
     score = 0.0
 
     p, c = match_info[:2]
-    if not raw_weights:
-        weights = [1] * c
-    else:
-        weights = [0] * c
-        for i, weight in enumerate(raw_weights):
-            weights[i] = weight
+    weights = get_weights(c, raw_weights)
 
+    # matchinfo X value corresponds to, for each phrase in the search query, a
+    # list of 3 values for each column in the search table.
+    # So if we have a two-phrase search query and three columns of data, the
+    # following would be the layout:
+    # p0 : c0=[0, 1, 2],   c1=[3, 4, 5],    c2=[6, 7, 8]
+    # p1 : c0=[9, 10, 11], c1=[12, 13, 14], c2=[15, 16, 17]
     for phrase_num in range(p):
         phrase_info_idx = 2 + (phrase_num * c * 3)
         for col_num in range(c):
@@ -1148,9 +1135,15 @@ def rank(raw_match_info, *raw_weights):
                 continue
 
             col_idx = phrase_info_idx + (col_num * 3)
-            x1, x2 = match_info[col_idx:col_idx + 2]
-            if x1 > 0:
-                score += weight * (float(x1) / x2)
+
+            # The idea is that we count the number of times the phrase appears
+            # in this column of the current row, compared to how many times it
+            # appears in this column across all rows. The ratio of these values
+            # provides a rough way to score based on "high value" terms.
+            row_hits = match_info[col_idx]
+            all_rows_hits = match_info[col_idx + 1]
+            if row_hits > 0:
+                score += weight * (float(row_hits) / all_rows_hits)
 
     return -score
 
@@ -1176,12 +1169,7 @@ def bm25(raw_match_info, *args):
     L_O = A_O + col_count
     X_O = L_O + col_count
 
-    if not args:
-        weights = [1] * col_count
-    else:
-        weights = [0] * col_count
-        for i, weight in enumerate(args):
-            weights[i] = args[i]
+    weights = get_weights(col_count, args)
 
     for i in range(term_count):
         for j in range(col_count):
@@ -1196,7 +1184,7 @@ def bm25(raw_match_info, *args):
             else:
                 D = 1 - B + (B * (doc_length / avg_length))
 
-            x = X_O + (3 * j * (i + 1))
+            x = X_O + (3 * (j + i * col_count))
             term_frequency = float(match_info[x])
             docs_with_term = float(match_info[x + 2])
 
